/*
 *  Copyright (C) 2010  Regents of the University of Michigan
 *
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "MemoryInfo.h"

String & MemoryInfo(double bytes)
{
    static String info;

    if (bytes < 1024)
        return info = "<1.0 kb";

    if (bytes < 1024. * 1024.)
        info.printf("%.1f kb", (bytes + 1023) / 1024.);
    else if (bytes < 1024. * 1024. * 1024.)
        info.printf("%.1f mb", (bytes + 1024. * 1024. - 1) / (1024. * 1024.));
    else if (bytes < 1024. * 1024. * 1024. * 1024.)
        info.printf("%.1f gb", bytes / (1024. * 1024. * 1024.));
    else
        info.printf("%.1f tb", bytes / (1024. * 1024. * 1024. * 1024.));

    return info;
}
